-----------------------------------------------------------------------
---- 2D Tile Terrain Generator originally by Wade Kallhoff
---- Translated to a Lua library by Jared "Nergal" Hewitt
---
--- Wade Kallhoff - http://mavdisk.mnsu.edu/kallhw/games.html
--- Original Source: http://mavdisk.mnsu.edu/kallhw/world/world.txt
--
-- uses a dynamically generated height-map to create realistic terrain
-----------------------------------------------------------------------

LuaBit = require("LuaBit")
SimplexNoise = require("SimplexNoise")

local table = table
local random = math.random
local round = function(num, idp)
  local mult = 10^(idp or 0)
  return math.floor(num * mult + 0.5) / mult
end

MapGen = {}

-------- Primary Functions --------

local Generate = function(w, h, tileScale, noiseScale)
	--defaults
	tileScale = tileScale or 10
	noiseScale = noiseScale or 0.1

	--argument checking
	assert(type(w) == "number", "MapGen.Generate: First argument expects number type")
	assert(type(h) == "number", "MapGen.Generate: Second argument expects number type")
	assert(type(tileScale) == "number", "MapGen.Generate: Third argument expects number type")
	assert(type(noiseScale) == "number", "MapGen.Generate: Fourth argument expects number type")

	local map = {}

    map = MapGen.Sub.initializeMap(w, h) --create a map object
    map = MapGen.Sub.generateHeightmap(map, w, h, noiseScale)

    map = MapGen.Sub.generateLakes(map, w, h)
    map = MapGen.Sub.generateMountains(map, w, h)
    map = MapGen.Sub.addRivers(map)
    map = MapGen.Sub.generateForest(map, w, h)
	
	map.tileScale = tileScale
	return map
end

local Draw = function(map) 
	MapGen.Sub.display(map)
end

MapGen.Generate = Generate
MapGen.Draw = Draw

-------- Sub Functions --------

-- Map Generation

local initializeMap = function(w, h)
	local map = {}
	map.mwidth = w
	map.mheight = h
	for x = 1, h, 1 do
		map[x] = {}
		for y = 1, w, 1 do
			map[x][y] = MapGen.Sub.tinit(x,y) --sets all of the map tiles initially to the same height and type
		end
	end
	return map
end

local generateHeightmap = function(map, w, h, noiseScale)
	for y = 1, h, 1 do
		for x = 1, w, 1 do
			--noiseVal = SimplexNoise.Noise2D(x*(noiseScale+random(0.5, 1)), y*(noiseScale+random(0.5, 1)))
			noiseVal = SimplexNoise.Noise2D(x*noiseScale, y*noiseScale)
			map[x][y].theight = noiseVal*255
			if map[x][y].theight < 20 then
				map[x][y].theight = 20
			end
		end
	end
	return map
end

local generateLakes = function(map, w, h)
	for y = 1, h, 1 do
		for x = 1, w, 1 do
			if map[x][y].theight < 100 then
				map[x][y].ttype = 1
			elseif map[x][y].theight < 110 then
				map[x][y].ttype = 4
			end
		end
	end
	return map
end

local addRivers = function(map, num)
	--defaults
	num = num or 20
	
	visited = {}
	for i = 0, num, 1 do
		local x = round(random(1,99))
		local y = round(random(1,99))
		map, visited = MapGen.Sub.generateRiver(map, visited, x, y); 
	end
	return map
end

local generateMountains = function(map, w, h)
	for y = 1, h, 1 do
		for x = 1, w, 1 do
			if map[x][y].theight > 150 then
				map[x][y].ttype = 2
			end
			if map[x][y].theight > 170 then
				map[x][y].ttype = 3
			end
		end
	end
	return map
end

local generateForest = function(map, w, h)
	for y = 1, h, 1 do
		for x = 1, w, 1 do
			chance = round(random(0, map[x][y].theight))
			if chance < 15 and map[x][y].ttype ~= 1 and map[x][y].ttype ~= 2 and map[x][y].ttype ~= 3 and map[x][y].ttype ~= 4 then
				map[x][y].ttype = 5
			end
		end
	end
	return map
end

local generateRiver = function(map, visited, x, y) --complicated and probably poor way to generate rivers, but it makes good rivers that follow the height. 
    lowest = map[x][y]

    for i = 0, 100, 1 do
		lowest.ttype = 1
		move = round(random(0,100))
		if lowest.x ~= 1 and lowest.x ~= 99 and lowest.y ~= 1 and lowest.y ~= 99 then
			up = map[lowest.x][lowest.y-1].theight;
			down = map[lowest.x][lowest.y+1].theight;
			left = map[lowest.x-1][lowest.y].theight;
			right = map[lowest.x+1][lowest.y].theight;
			if move > 80 then
				rand = round(random(0,3))
				if rand == 0 then
					lowest = map[lowest.x+1][lowest.y]
				elseif rand == 1 then
					lowest = map[lowest.x-1][lowest.y]
				elseif rand == 2 then
					lowest = map[lowest.x][lowest.y+1]
				elseif rand == 3 then
					lowest = map[lowest.x][lowest.y-1]
				end
			elseif up < down and up < left and up < right then
				lowest = map[lowest.x][lowest.y-1]
			elseif down < up and down < left and down < right then
				lowest = map[lowest.x][lowest.y+1]
			elseif left < down and left < up and left < right then
				lowest = map[lowest.x-1][lowest.y]
			elseif right < down and right < left and right < up then
				lowest = map[lowest.x+1][lowest.y]
			end

			if not visited[lowest] then
				lowest.ttype = 1
				visited[lowest] = "visited"
			else
				rand = round(random(0,3))
				if rand == 0 then
					lowest = map[lowest.x+1][lowest.y]
				elseif rand == 1 then
					lowest = map[lowest.x-1][lowest.y]
				elseif rand == 2 then
					lowest = map[lowest.x][lowest.y+1]
				elseif rand == 3 then
					lowest = map[lowest.x][lowest.y-1]
				end
				visited[lowest] = "visited"
				lowest.ttype = 1
			end
		end
    end
	return map, visited
end

local mdisplay = function(map)
	for x = 1, map.mwidth, 1 do
		for y = 1, map.mheight, 1 do
			map[x][y].display(map[x][y].ttype, map[x][y].theight, map.tileScale, x, y)
		end
    end
end

-- Tile functions

-- Function to construct our tile and push it into a table
local tinit = function(tx, ty, t, h)
	--defaults
	t = t or 0
	h = h or 0

	local tile = {}
	tile.x = tx --x location of the tile
	tile.y = ty --y location of the tile
	tile.ttype = t --terrain type of the tile water/sand/grass etc
	tile.theight = h --terrain height
	tile.display = MapGen.Sub.tdisplay

	return tile
end

--Gets assigned to tile.display so that it can be called directly from the tile's table
local tdisplay = function(ttype, theight, tileScale, x, y)
	if ttype == 0 then
		love.graphics.setColor(0,255,0,theight)
	elseif ttype == 1 then
		love.graphics.setColor(0,0,255,theight)
	elseif ttype == 2 then
		love.graphics.setColor(100,100,100,theight)
	elseif ttype == 3 then
		love.graphics.setColor(220,220,220,theight)
	elseif ttype == 4 then
		love.graphics.setColor(237,255,49,theight)
	elseif ttype == 5 then
		love.graphics.setColor(19,64,7)
	else
		love.graphics.setColor(0,0,0)
	end
	
	local tilex = (x*tileScale)-(tileScale-1)
	local tiley = (y*tileScale)-(tileScale-1)

	love.graphics.rectangle('fill', tilex, tiley, tileScale, tileScale)
end

MapGen.Sub = {}
MapGen.Sub.initializeMap = initializeMap
MapGen.Sub.generateHeightmap = generateHeightmap
MapGen.Sub.generateLakes = generateLakes
MapGen.Sub.addRivers = addRivers
MapGen.Sub.generateMountains = generateMountains
MapGen.Sub.generateForest = generateForest
MapGen.Sub.generateRiver = generateRiver
MapGen.Sub.display = mdisplay
MapGen.Sub.tdisplay = tdisplay
MapGen.Sub.tinit = tinit

return MapGen